/* ************************************************************************** *
 *      Copyright © IPSL/LSCE, XMLIOServer, Avril 2010 - Octobre 2011         *
 * ************************************************************************** */
 
 /**
 * \file    mpi_interface.hpp
 * \brief   Gestion des communications MPI via une surcouche interne (implémentation).
 * \author  Hervé Ozdoba
 * \version 0.4
 * \date    28 Juin 2011
 */
 
// XMLIOServer headers
#include "mpi_interface.hpp"

// /////////////////////////////// Définitions ////////////////////////////// //

namespace xmlioserver {
namespace comm {

   // ---------------------- Initialisation & Finalisation ---------------------

   void CMPIManager::Initialise(int * _argc, char *** _argv)
   {
      int flag = 0;
      if (MPI_Initialized(&flag) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Initialise(arc, argv)", << " MPI Error !");
      if (!flag)
      {
         if (MPI_Init(_argc, _argv) != MPI_SUCCESS)
            XIOS_ERROR("CMPIManager::Initialise(arc, argv)", << " MPI Error !");
      }

   }
   
   void CMPIManager::Finalize(void)
   {
      if (MPI_Finalize() != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Finalize(void)", << " MPI Error !");
   }
   
   // ------------------------------ Communicateurs ----------------------------
   
   int CMPIManager::GetCommRank(MPI_Comm _comm)
   {
      int rank = 0;
      if (MPI_Comm_rank(_comm, &rank) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::GetCommRank(comm)", << " MPI Error !");
      return (rank);
   }
   
   int CMPIManager::GetCommSize(MPI_Comm _comm)
   {
      int size = 0;
      if (MPI_Comm_size(_comm, &size) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::GetCommSize(comm)", << " MPI Error !");
      return (size);
   }
   
   MPI_Comm CMPIManager::CreateComm(MPI_Group _group, MPI_Comm _pcomm)
   {
      MPI_Comm commu;      
      if (MPI_Comm_create(_pcomm, _group, &commu) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::CreateComm(group, pcomm)", << " MPI Error !");
      return (commu);
   }
   
   MPI_Comm CMPIManager::GetCommWorld(void)
   {
      return (MPI_COMM_WORLD); 
   }
   
   // ---------------------------------- Autre ---------------------------------
         
   void CMPIManager::Barrier(MPI_Comm _comm)
   {
      if (MPI_Barrier(_comm) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Barrier(comm)", << " MPI Error !");
   }

   // --------------------------------- Groupes --------------------------------
         
   MPI_Group CMPIManager::GetGroupWorld(void)
   {
      MPI_Group group = 0;
      if (MPI_Comm_group(CMPIManager::GetCommWorld(), &group) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::GetGroupWorld()", << " MPI Error !");
      return (group);
   }
   
   MPI_Group CMPIManager::CreateSubGroup(MPI_Group _pgroup, const std::vector<int> & _ranks)
   {
      MPI_Group group = 0;
      if (MPI_Group_incl(_pgroup, _ranks.size(), const_cast<int*>(&(_ranks[0])), &group) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::CreateSubGroup(pgroup, ranks)", << " MPI Error !");
      return (group);
   }
   
   MPI_Group CMPIManager::CreateSubGroup
      (MPI_Group _pgroup, int _min_rank, int _max_rank, int _intval)
   {
      std::vector<int> ranks;
      for (int i = _min_rank; i <= _max_rank; i += _intval)
         ranks.push_back(i);
      return (CMPIManager::CreateSubGroup(_pgroup, ranks));
   }

   // ----------------------------------- Tests --------------------------------
         
   bool CMPIManager::IsMaster(MPI_Comm _comm)
   {
      return (CMPIManager::GetCommRank(_comm) == 0); 
   }
   
   bool CMPIManager::IsRank(int _rank, MPI_Comm _comm)
   {
      return (CMPIManager::GetCommRank(_comm) == _rank); 
   }

   // --------------------------- Communication simple -------------------------
         
   void CMPIManager::Send (MPI_Comm _comm, int _dest_rank, char * _data,
                           std::size_t _size, MPI_Request & _request)
   {
      int nsize = _size;    
      if (MPI_Issend(_data, nsize, MPI_CHAR, _dest_rank, 0, _comm, &_request) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Send (comm, dest_rank, data, size, request)", << " MPI Error !");
   }
   
   void CMPIManager::Wait (MPI_Request & _request)
   {
      MPI_Status status;
      if (MPI_Wait(&_request, &status) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Wait (request)", << " MPI Error !");
   }
   
   bool CMPIManager::Test (MPI_Request & _request)
   {
      MPI_Status status;
      int flag = 0;
      if (MPI_Test(&_request, &flag, &status) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Test (request)", << " MPI Error !");
      return (flag);
   }

   bool CMPIManager::HasReceivedData(MPI_Comm _comm, int _src_rank)
   {
      MPI_Status status;
      int flag = 0;
      if (MPI_Iprobe(_src_rank, MPI_ANY_TAG, _comm, &flag, &status) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::HasReceivedData (comm, rank)", << " MPI Error !");
      return (flag);
   }
   
   std::size_t CMPIManager::GetReceivedDataSize(MPI_Comm _comm, int _src_rank)
   {
      MPI_Status status;
      int flag = 0, size = 0;
      if (MPI_Iprobe(_src_rank, MPI_ANY_TAG, _comm, &flag, &status) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::getReceivedDataSize (comm, rank)", << " MPI Error !");
      if (!flag) return (0);  
      if (MPI_Get_count(&status, MPI_CHAR, &size) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::getReceivedDataSize (comm, rank)", << " MPI Error !");

      return (size);
   }
   
   void CMPIManager::Receive(MPI_Comm _comm, int _src_rank, char * _data)
   {
      MPI_Request request = 0;
      int size = CMPIManager::GetReceivedDataSize(_comm, _src_rank);
      if (MPI_Irecv(_data, size, MPI_CHAR, _src_rank, MPI_ANY_TAG, _comm, &request) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::Receive (comm, src_rank, data)", << " MPI Error !");
      CMPIManager::Wait (request); // Temporaire
   }
         
   // ------------------------- Communication 'complexe' -----------------------
         
   //void SendLinearBuffer(MPIComm comm, int dest_rank, CLinearBuffer & buff, MPIRequest & request);
   //void ReceiveLinearBuffer(MPIComm comm, int src_rank, CLinearBuffer & buff);
   //boost::shared_ptr<CLinearBuffer> ReceiveLinearBuffer(MPIComm comm, int src_rank);
   //void ReceiveCircularBuffer(MPIComm comm, int src_rank, CCircularBuffer & buff);

   // ---------------------- Mémoire (non fonctionnel ....) --------------------
         
   void CMPIManager::AllocMemory(void * _data, std::size_t _size)
   {
      if (MPI_Alloc_mem(sizeof(char) * _size, MPI_INFO_NULL, _data) != MPI_SUCCESS)
         XIOS_ERROR("CMPIManager::AllocMem(data, size)", << " MPI Error !");
   }
   
   void CMPIManager::FreeMemory (void * _data)
   {
      MPI_Free_mem(_data);
   }

} // namespace comm
} // namespace xmlioserver

